/*****************************************************************************/
/* tokenize.c 								     */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Standalone library for basic text processing.			     */
/* (c) 2001, 2004, Pal, A. (apal@szofi.elte.hu). 			     */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* See function prototypes and the usage of the functions in tokenize.h	     */
/*****************************************************************************/

#include <stdio.h>
#include <memory.h>
#include <string.h>
#include <stdlib.h>

#include "tokenize.h"

/*****************************************************************************/

void remove_newlines_and_comments(char *buff)
{
 int k;
 while ( *buff )
  {	if ( *buff=='#' )	*buff=0;
	else 
	 {	for ( k=0 ; buff[k]==10 || buff[k]==13 ; )	k++;
		if ( k )	memmove(buff,buff+k,strlen(buff)+1-k);
		else		buff++;
	 }
  }
	
}

void remove_spaces_and_comments(char *buff)
{
 int k;
 while ( *buff )
  {	if ( *buff=='#' )	*buff=0;
	else 
	 {	for ( k=0 ; buff[k]==9 || buff[k]==32 || buff[k]==10 || buff[k]==13 ; )	k++;
		if ( k )	memmove(buff,buff+k,strlen(buff)+1-k);
		else		buff++;
	 }
  }
	
}

void remove_spaces(char *buff)
{
 int k;
 while ( *buff )
  {	for ( k=0 ; buff[k]==9 || buff[k]==32 || buff[k]==10 || buff[k]==13 ; )	k++;
	if ( k )	memmove(buff,buff+k,strlen(buff)+1-k);
	else		buff++;
  }
}

/*****************************************************************************/

void remove_quotes(char *buff)
{
 int k;
 while ( *buff )
  {	for ( k=0 ; buff[k]=='"' ; )	k++;
	if ( k )	memmove(buff,buff+k,strlen(buff)+1-k);
	else		buff++;
  }
	
}

/*****************************************************************************/

int char_is_space(int c)
{
 if ( c==32 || c==13 || c==10 || c==9 )	return(1);
 else					return(0);
}

/*****************************************************************************/

int tokenize_spaces(char *buff,char **tokens,int max)
{
 int	intoken,inquota,n;
 char 	**tsave;

 tsave=tokens;

 intoken=0,inquota=0;n=0;
 while ( *buff && n<max )
  {	if ( ( ! char_is_space(*buff) ) && ! intoken )
	 {	*tokens=buff;
		intoken=!0,inquota=0;n++;
		if ( *buff=='"' )	inquota=!0;
		tokens++,buff++;
	 }
	else if ( intoken && ( (char_is_space(*buff) && inquota) || (!char_is_space(*buff)) ) )
	 {	if ( *buff=='"' )	inquota=!inquota;
		buff++;
	 }
	else if ( intoken && ! inquota && char_is_space(*buff) )
	 {	*buff=0,buff++;
		intoken=0;
	 }
	else	buff++;
  };
 *tokens=NULL;

 while ( *tsave != NULL )
  {	remove_quotes(*tsave);
	tsave++;
  };

 return(n);
}

char **tokenize_spaces_dyn(char *buff)
{
 int	intoken,inquota,i,n,nm;
 char 	**rtokens;

 nm=16;
 rtokens=(char **)malloc(sizeof(char *)*nm);

 intoken=0,inquota=0;n=0;
 while ( *buff )
  {	if ( ( ! char_is_space(*buff) ) && ! intoken )
	 {	rtokens[n]=buff;
		intoken=!0,inquota=0;n++;
		if ( *buff=='"' )	inquota=!0;
		buff++;
		if ( n>=nm-1 )
		 {	nm+=16;
			rtokens=(char **)realloc(rtokens,sizeof(char *)*nm);
		 }
	 }
	else if ( intoken && ( (char_is_space(*buff) && inquota) || (!char_is_space(*buff)) ) )
	 {	if ( *buff=='"' )	inquota=!inquota;
		buff++;
	 }
	else if ( intoken && ! inquota && char_is_space(*buff) )
	 {	*buff=0,buff++;
		intoken=0;
	 }
	else	buff++;
  };

 rtokens[n]=NULL;

 for ( i=0 ; i<n ; i++ )
  {	remove_quotes(rtokens[i]);	}

 return(rtokens);
}

/*****************************************************************************/

int tokenize_char(char *buff,char **tokens,int tchar,int max)
{
 int	n;

 if ( *buff==0 )
  {	*tokens=NULL;return(0);		}
	
 n=1;*tokens=buff,tokens++;
 while ( *buff && n<max )
  {	if ( *buff != tchar )	buff++;
	else			*buff=0,buff++,*tokens=buff,tokens++,n++;
  };
 *tokens=NULL;
 return(n);
}

char **tokenize_char_dyn(char *buff,int tchar)
{
 int	n;
 char	**tokens;

 if ( buff==NULL )
  {	return(NULL);				}

 tokens=(char **)malloc(sizeof(char *));
 if ( *buff==0 )
  {	*tokens=NULL;return(tokens);		}
	
 n=0;tokens[n]=buff,n++;
 while ( *buff )
  {	if ( *buff != tchar )	buff++;
	else
	 {	*buff=0,buff++;
		tokens=(char **)realloc(tokens,sizeof(char *)*(n+1));
		tokens[n]=buff,n++;
	 }
  };
 tokens=(char **)realloc(tokens,sizeof(char *)*(n+1));
 tokens[n]=NULL;
 return(tokens);
}

/*****************************************************************************/

#define		token_add(tokens,ntoken) 				\
 do 									\
  {	tokens=(char **)realloc(tokens,sizeof(char *)*(ntoken+1));	\
  } while(0)

char **tokenize_shell_dyn(char *buff)
{
 int	intoken,inquota,is_esc;
 char 	**tokens;
 int	ntoken;

 char	*copy;

 tokens=NULL;
 ntoken=0;

 copy=buff;

 intoken=0;
 inquota=0;

 while ( *buff )
  {	if ( *buff=='\\' && *(buff+1)==10 )
	 {	buff+=2;
		continue;
	 }
	else if ( *buff=='\\' )
 	 {	is_esc=1;
		buff++;
	 }
	else
		is_esc=0;

	/* fprintf(stderr,">> [%c] inquota=%d is_esc=%d intoken=%d\n",*buff,inquota,is_esc,intoken); */

	if ( char_is_space(*buff) && !inquota && !is_esc && intoken )
	 {	buff++;
		*copy=0;
		copy++;
		intoken=0;
	 }
	else if ( char_is_space(*buff) && !intoken && !is_esc )
	 {	buff++;
		*copy=0;
		copy++;
	 }
	else if ( !is_esc && !inquota && (*buff=='\"' || *buff=='\'') )
	 {	inquota=*buff;
		buff++;
		if ( ! intoken )
		 {	token_add(tokens,ntoken);
			tokens[ntoken]=copy;
			ntoken++;
			intoken=1;
		 }
	 }
	else if ( !is_esc && inquota && (*buff)==inquota )
	 {	inquota=0;
		buff++;
	 }
	else if ( !intoken )
	 {	token_add(tokens,ntoken);
		tokens[ntoken]=copy;
		ntoken++;
		intoken=1;
		*copy=*buff;
		copy++;buff++;
	 }
	else
	 {	*copy=*buff;
		copy++;buff++;
	 } 
	
  };

 *copy=0;
 token_add(tokens,ntoken);
 tokens[ntoken]=NULL;

 return(tokens);
}

/*****************************************************************************/
             

